# Copyright 2005, Cadalog, Inc.

#-----------------------------------------------------------------------------

require 'Cadalog/Win32API'
require 'sketchup.rb'

#-----------------------------------------------------------------------------


class BeamTool

# This is the standard Ruby initialize method that is called when you create
# a new object.
def initialize
    @ip1 = nil
    @ip2 = nil
    @xdown = 0
    @ydown = 0
    @beam_length = -1
    dll_name = Sketchup.find_support_file ("CadalogParts.dll", "Plugins/Cadalog")
print "dll_name is '"
print dll_name
print "'\n"
    @dll_Load = Win32API.new(dll_name,"AddOnLoad",[],'L')
    @dll_Unload = Win32API.new(dll_name,"AddOnUnload",[],'L')
    @dll_Pick = Win32API.new(dll_name,"AddOnPick",[],'L')
    @dll_Length = Win32API.new(dll_name,"AddOnLength","I",'P')
    @dll_Profile = Win32API.new(dll_name,"AddOnProfile","I",'P')
    @dll_ProfileLines = Win32API.new(dll_name,"AddOnProfileLines","I",'P')
end

# The activate method is called by SketchUp when the tool is first selected.
# it is a good place to put most of your initialization
def activate
    # The Sketchup::InputPoint class is used to get 3D points from screen
    # positions.  It uses the SketchUp inferencing code.
    # In this tool, we will have two points for the endpoints of the line.
    @ip1 = Sketchup::InputPoint.new
    @ip2 = Sketchup::InputPoint.new
    @ip = Sketchup::InputPoint.new
    @drawn = false

	if (!@dll_loaded)
		@dll_Load.call()
	end
    @dll_loaded = true

    @picked = @dll_Pick.call()
	if (! @picked)
		# user cancelled, cancel everything
		@beam_length = -1
		return
	end

    # This sets the label for the VCB
    #Sketchup::set_status_text "Length", SB_VCB_LABEL

	units = Sketchup.active_model.options["UnitsOptions"]["LengthUnit"]
	@beam_length = @dll_Length.call(units).to_f
	str = ""
	@profile_lines = Array.new
	if (@beam_length > 0)
		# construct the list of points that make up the temporary profile
		# we look only at line segments, and assume that the line segments
		# are connected one to another in the order they are sent to us
		str = @dll_ProfileLines.call(units)
		ops = str.split (/;/)
		ops.each { |op|
			attr = Hash.new
			list = op.split (/\s+/)
			list.each { |a| cmd,val = a.split(/=/); attr[cmd] = val }
			if (attr['cmd'] == 'l')
				# line
				if (@profile_lines.size == 0)
					# if this is the first line segment, start at its starting point
					@profile_lines.push(Geom::Point3d.new(attr['sx'].to_f, attr['sy'].to_f, 0.0))
				end
				@profile_lines.push(Geom::Point3d.new(attr['ex'].to_f, attr['ey'].to_f, 0.0))
			end
		}
	end

    self.reset(nil)
end


# deactivate is called when the tool is deactivated because
# a different tool was selected
def deactivate(view)
    view.invalidate if @drawn
    if (@dll_loaded)
    	@dll_Unload.call()
    end
    @dll_loaded = false
end


# The onMouseMove method is called whenever the user moves the mouse.
# because it is called so often, it is important to try to make it efficient.
# In a lot of tools, your main interaction will occur in this method.
def onMouseMove(flags, x, y, view)
    if( @state == 0 )
        # We are getting the first end of the line.  Call the pick method
        # on the InputPoint to get a 3D position from the 2D screen position
        # that is bassed as an argument to this method.
        @ip.pick view, x, y
        if( @ip != @ip1 )
            # if the point has changed from the last one we got, then
            # see if we need to display the point.  We need to display it
            # if it has a display representation or if the previous point
            # was displayed.  The invalidate method on the view is used
            # to tell the view that something has changed so that you need
            # to refresh the view.
            ##view.invalidate if( @ip.display? or @ip1.display? )
            view.invalidate
            @ip1.copy! @ip

            # set the tooltip that should be displayed to this point
            view.tooltip = @ip1.tooltip
        end
    elsif( @state == 1 )
        # Getting the second end of the line
        # If you pass in another InputPoint on the pick method of InputPoint
        # it uses that second point to do additional inferencing such as
        # parallel to an axis.
        @ip2.pick view, x, y, @ip1
        view.tooltip = @ip2.tooltip if( @ip2.valid? )
        view.invalidate

        # Update the length displayed in the VCB
       if( @ip2.valid? )
            length = @ip1.position.distance(@ip2.position)
            #Sketchup::set_status_text length.to_s, SB_VCB_VALUE
        end

        # Check to see if the mouse was moved far enough to create a line.
        # This is used so that you can create a line by either draggin
        # or doing click-move-click
        if( (x-@xdown).abs > 10 || (y-@ydown).abs > 10 )
            @dragging = true
        end
    end
end

# The onLButtonDOwn method is called when the user presses the left mouse button.
def onLButtonDown(flags, x, y, view)
	if (@beam_length <= 0)
		return
	end
    # When the user clicks the first time, we switch to getting the
    # second point.  When they click a second time we create the line
    if( @state == 0 )
        @ip1.pick view, x, y
        if( @ip1.valid? )
            @state = 1
            Sketchup::set_status_text "Select second end", SB_PROMPT
            @xdown = x
            @ydown = y
        end
    elsif( @state == 1 )
        # create the line on the second click
        if( @ip2.valid? )
            self.create_geometry(@ip1.position, @ip2.position,view)
            self.reset(view)
        end
    end

    # Clear any inference lock
    view.lock_inference
end

# The onLButtonUp method is called when the user releases the left mouse button.
def onLButtonUp(flags, x, y, view)
	if (@beam_length <= 0)
		return
	end
    # If we are doing a drag, then create the line on the mouse up event
    if( @dragging && @ip2.valid? )
        self.create_geometry(@ip1.position, @ip2.position,view)
        self.reset(view)
    end
end

# onKeyDown is called when the user presses a key on the keyboard.
# We are checking it here to see if the user pressed the shift key
# so that we can do inference locking
def onKeyDown(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
       @shift_down_time = Time.now

        # if we already have an inference lock, then unlock it
        if( view.inference_locked? )
            # calling lock_inference with no arguments actually unlocks
            view.lock_inference
        elsif( @state == 0 && @ip1.valid? )
            view.lock_inference @ip1
        elsif( @state == 1 && @ip2.valid? )
            view.lock_inference @ip2, @ip1
        end
    end
end

# onKeyUp is called when the user releases the key
# We use this to unlock the interence
# If the user holds down the shift key for more than 1/2 second, then we
# unlock the inference on the release.  Otherwise, the user presses shift
# once to lock and a second time to unlock.
def onKeyUp(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY &&
        view.inference_locked? &&
        (Time.now - @shift_down_time) > 0.5 )
        view.lock_inference
    end
end

# The draw method is called whenever the view is refreshed.  It lets the
# tool draw any temporary geometry that it needs to.
def draw(view)
    if( @ip1.valid? )
        self.draw_geometry(@ip1, nil, view)
        if( @ip1.display? )
            @ip1.draw(view)
            @drawn = true
        end

        if( @ip2.valid? )
            @ip2.draw(view) if( @ip2.display? )

            # The set_color_from_line method determines what color
            # to use to draw a line based on its direction.  For example
            # red, green or blue.
            view.set_color_from_line(@ip1, @ip2)
            self.draw_geometry(@ip1, @ip2, view)
            @drawn = true
        end
    end
end

# onCancel is called when the user hits the escape key
def onCancel(flag, view)
    self.reset(view)
end


# The following methods are not directly called from SketchUp.  They are
# internal methods that are used to support the other methods in this class.

# Reset the tool back to its initial state
def reset(view)
    # This variable keeps track of which point we are currently getting
    @state = 0

    # Display a prompt on the status bar
    Sketchup::set_status_text("Select first end", SB_PROMPT)

    # clear the InputPoints
    @ip1.clear
    @ip2.clear

    if( view )
        view.tooltip = nil
        view.invalidate if @drawn
    end

    @drawn = false
    @dragging = false
end

# Create new permanent geometry when the user has selected two points.
def create_geometry(p1, p2, view)
	if (@beam_length <= 0)
		return
	end
	# make a new group to contain the constructed beam
	beam = view.model.entities.add_group
	# get the entities within the group
	beam_entities = beam.entities

	units = Sketchup.active_model.options["UnitsOptions"]["LengthUnit"]
	str = ""
	edges = Array.new
	@beam_length = @dll_Length.call(units).to_f

	if (@beam_length)
		str = @dll_Profile.call(units)
		ops = str.split (/;/)
		ops.each { |op|
			attr = Hash.new
			list = op.split (/\s+/)
			list.each { |a| cmd,val = a.split(/=/); attr[cmd] = val }
			if (attr['cmd'] == 'l')
				# line
				startPt = @final_xform * Geom::Point3d.new(attr['sx'].to_f, attr['sy'].to_f, 0.0)
				endPt = @final_xform * Geom::Point3d.new(attr['ex'].to_f, attr['ey'].to_f, 0.0)
				edges.push (beam_entities.add_line (startPt, endPt))
			end
			if (attr['cmd'] == 'a')
				# arc
				startAngle = attr['sa'].to_f
				endAngle = attr['ea'].to_f
				radius = attr['r'].to_f
				centerPt = @final_xform * Geom::Point3d.new(attr['cx'].to_f, attr['cy'].to_f, 0.0)
				xAxis = @final_xform * Geom::Vector3d.new(1.0, 0.0, 0.0)
				normal = @final_xform.zaxis
				edges.concat(beam_entities.add_arc (centerPt, xAxis, normal, radius, startAngle, endAngle))
			end
			if (attr['cmd'] == 'c')
				# circle
				centerPt = @final_xform * Geom::Point3d.new(attr['cx'].to_f, attr['cy'].to_f, 0.0)
				normal = @final_xform.zaxis
				radius = attr['r'].to_f
				edges.concat(beam_entities.add_circle (centerPt, normal, radius))
			end
		}
	end
    face = beam_entities.add_face(edges)
    face.pushpull @beam_length
end

# Draw the geometry as the part is dragged around
def draw_geometry(p1, p2, view)
	if (@beam_length <= 0)
		return
	end
    if( @state == 0 )
    	# if p1 is on a face, create a transform that places the profile on that face
    	# otherwise just place the profile on the point
		if( p1.face )
			@p1_xform = Geom::Transformation.new(p1.position, p1.face.normal)
		else
			@p1_xform = Geom::Transformation.new(p1.position)
		end
	end
	if( p2 )
		# rotate the p1 transform to line up with p2
		inverse_xform = @p1_xform.inverse
		inverse_p2 = inverse_xform * p2.position
		rotation_angle = Math.atan2(inverse_p2.y, inverse_p2.x) - Math.asin(1)
		rotation_xform = Geom::Transformation.rotation(Geom::Point3d.new, Geom::Vector3d.new(0,0,1), rotation_angle)
		@final_xform = @p1_xform * rotation_xform
	else
		@final_xform = @p1_xform
	end

    xlatedPts = Array.new
    @profile_lines.each { | pt | xlatedPts.push ( @final_xform * pt) }
    view.draw_polyline xlatedPts
end

end # class BeamTool


# Add some menu items to access this
if( not file_loaded?("beamtool.rb") )
    plugins_menu = UI.menu("Plugins")
    plugins_menu.add_item("Add Beam") { Sketchup.active_model.select_tool BeamTool.new }
end


#-----------------------------------------------------------------------------
file_loaded("beamtool.rb")
